"""Implementation of CLI modify command."""

import math

from itertools import chain

import numpy as np

from treem.morph import Morph
from treem.io import SWC
from treem.utils.geom import rotation


def modify(args):
    """Modifies selected parts of morphology reconstruction."""
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    morph = Morph(args.file)
    if args.ids:
        nodes = filter(lambda x: x.ident() in args.ids, morph.root.walk())
    else:
        sections = chain.from_iterable(x.sections() for x in morph.stems())
        nodes = chain(x[0] for x in sections)
    types = args.type if args.type else SWC.TYPES
    nodes = filter(lambda x: x.type() in types, nodes)
    if args.order:
        nodes = filter(lambda x: x.order() in args.order, nodes)
    if args.breadth:
        nodes = filter(lambda x: x.breadth() in args.breadth, nodes)
    nodes = list(nodes)

    if args.scale_radius:
        for node in nodes:
            sec = list(node.section())
            radii = morph.radii(sec)
            radii *= args.scale_radius

    if args.scale:
        for node in nodes:
            sec = list(node.section())
            head = sec[0].coord().copy()
            tail = sec[-1].coord().copy()
            coords = morph.coords(sec)
            coords *= np.array(args.scale)
            shift = head - sec[0].coord()
            coords += shift
            for child in sec[-1].siblings:
                shift = sec[-1].coord() - tail
                morph.translate(shift, child)

    if args.seed:
        np.random.seed(args.seed)

    if args.jitter:
        for node in nodes:
            sec = list(node.section())
            if len(sec) > 1:
                head = sec[0].coord().copy()
                tail = sec[-1].coord().copy()
                length = morph.length(sec[1:])
                coords = morph.coords(sec)
                if not args.sec:
                    rnd = np.random.uniform(-1, 1, np.shape(coords))
                    coords += args.jitter * rnd
                else:
                    xlen = 0
                    rnd = np.random.uniform(-1, 1, 3)
                    for node in sec:
                        xlen += node.length()
                        vec = args.jitter * rnd * xlen / length
                        morph.move(vec, node)
                scale = length / morph.length(sec[1:])
                coords *= scale
                shift = head - sec[0].coord()
                coords += shift
                for child in sec[-1].siblings:
                    shift = sec[-1].coord() - tail
                    morph.translate(shift, child)

    if args.twist:
        for node in nodes:
            axis = node.coord() - node.parent.coord()
            angle = args.twist*np.random.uniform(-1, 1)*math.pi/180
            morph.rotate(axis, angle, node)

    if args.stretch:
        for node in nodes:
            sec = list(node.section())
            if len(sec) > 1:
                head = sec[0].coord().copy()
                tail = sec[-1].coord().copy()
                length = morph.length(sec[1:])
                coords = morph.coords(sec)
                vdir = coords.mean(axis=0) - head
                vdir /= np.linalg.norm(vdir)
                for secnode in sec[1:]:
                    coord = secnode.coord()
                    coord += vdir * np.linalg.norm(coord - head) * args.stretch
                scale = length / morph.length(sec[1:])
                coords *= scale
                shift = head - sec[0].coord()
                coords += shift
                for child in sec[-1].siblings:
                    shift = sec[-1].coord() - tail
                    morph.translate(shift, child)

    if args.smooth:
        for node in nodes:
            sec = list(node.section())
            if len(sec) > 2:
                head = sec[0].coord().copy()
                tail = sec[-1].coord().copy()
                length = morph.length(sec[1:])
                coords = morph.coords(sec)
                for i in range(args.smooth):  # pylint: disable=unused-variable
                    for secnode in sec[-1::-1]:
                        coord = secnode.coord()
                        coord += secnode.parent.coord()
                    scale = length / morph.length(sec[1:])
                    coords *= scale
                    shift = head - sec[0].coord()
                    coords += shift
                for child in sec[-1].siblings:
                    shift = sec[-1].coord() - tail
                    morph.translate(shift, child)

    if args.swap:
        np.random.shuffle(nodes)
        for node1, node2 in zip(nodes[:-1:2], nodes[1::2]):
            parent1 = node1.parent
            parent2 = node2.parent
            dir1 = node1.coord() - parent1.coord()
            dir2 = node2.coord() - parent2.coord()
            tree1 = morph.copy(node1)
            tree2 = morph.copy(node2)
            coord1 = tree1.root.coord().copy()
            coord2 = tree2.root.coord().copy()
            tree1.translate(coord2 - coord1)
            axis, angle = rotation(dir1, dir2)
            tree1.rotate(axis, angle)
            tree2.translate(coord1 - coord2)
            axis, angle = rotation(dir2, dir1)
            tree2.rotate(axis, angle)
            morph.graft(tree1, parent2)
            morph.graft(tree2, parent1)
            morph.prune(node1)
            morph.prune(node2)

    if args.prune and not args.swap:
        for node in nodes:
            morph.prune(node)

    if args.prune or args.swap:
        morph = Morph(data=morph.data)

    morph.save(args.out)



"""Utilities for manipulating geometry of morphology reconstructions."""

import math
import numpy as np

from treem.io import SWC

# pylint: disable=invalid-name
# pylint: disable=too-many-locals


def rotation_matrix(axis, angle):
    """Computes rotation matrix for 3D manipulations.

    Args:
        axis (float[3]): rotation axis.
        angle (float): rotation angle in radians.

    Returns:
        rotation matrix (NumPy ndarray).
    """
    axis = np.asarray(axis)
    axis = axis / math.sqrt(np.dot(axis, axis))
    a = math.cos(angle / 2.0)
    b, c, d = -axis * math.sin(angle / 2.0)
    aa, bb, cc, dd = a*a, b*b, c*c, d*d
    bc, ad, ac, ab, bd, cd = b*c, a*d, a*c, a*b, b*d, c*d
    return np.array([[aa + bb - cc - dd, 2*(bc + ad), 2*(bd - ac)],
                     [2*(bc - ad), aa + cc - bb - dd, 2*(cd + ab)],
                     [2*(bd + ac), 2*(cd - ab), aa + dd - bb - cc]])


def norm(vec):
    """Returns magnitude of a 3D vector."""
    x, y, z = vec
    return math.sqrt(x*x + y*y + z*z)


def angle_between(u, v):
    """Returns angle in radians between two 3D vectors (float[3])."""
    un, vn, = norm(u), norm(v)
    angle = math.acos(np.dot(u, v) / (un*vn)) if un > 1e-6 and vn > 1e-6 else 0
    return angle


def rotation(u, v):
    """Computes rotation axis and angle for two 3D vectors (float[3]).

    Returns:
        axis (NumPy ndarray[3]), angle (float)

        rotation axis, rotation angle in radians.
    """
    angle = angle_between(u, v)
    if math.isclose(math.sin(angle), 0, abs_tol=1e-6):
        dv = np.ones(3)*1e-6
        u, v = u + dv, v - dv
    axis = np.cross(u, v)
    axis = axis / norm(axis)
    return axis, angle


def repair_branch(cmorph, cut, rmorph, rep, force=False):
    """Attempts to extend cut neurite using intact branch.

    Args:
        cmorph (treem.Morph): cut morphology.
        cut (treem.Node): cut node, from cmorph.
        rmorph (treem.Morph): repair morphology.
        rep (treem.Node): undamaged branch start node, from rmorph.
        force (bool): force repair if branch is too short.

    Returns:
        True if repaired.
    """
    done = 0
    cutsec = list(reversed(list(cut.section(reverse=True))))
    repsec = list(rep.section())
    cutlen = cmorph.length(cutsec)
    replen = rmorph.length(repsec)
    target = cut
    if replen > cutlen:
        for node in repsec[-1::-1]:
            if rmorph.length(node.section()) > replen - cutlen:
                break
        source = node  # pylint: disable=undefined-loop-variable
    elif rep.breadth() > 1 or force:
        source = rep
    else:
        source = None
    if source:
        tree = rmorph.copy(source)
        scale_z = -1
        scale_r = cmorph.radii(cutsec).mean() / rmorph.radii(repsec).mean()
        tree.data[:, SWC.XYZR] *= np.array([1, 1, scale_z, scale_r])
        u = np.mean(tree.data[:, SWC.XYZ], axis=0) - tree.root.coord()
        v = target.coord() - cmorph.root.coord()
        axis, angle = rotation(u, v)
        tree.rotate(axis, angle)
        shift = (target.coord() - tree.root.coord() +
                 target.coord() - target.parent.coord())
        tree.translate(shift)
        cmorph.graft(tree, target)
        done = 1
    return done


def sample(points, num):
    """Samples points with fixed resolution using linear interpolation.

    Args:
        points (NumPy ndarray[4]): array of 4D data points (x,y,z,r).
        num (int): sample size.

    Returns:
        array of 4D points (NumPy ndarray[4]).
    """
    num = num if num > 1 else 2
    links = [norm(p[0:3] - q[0:3]) for p, q in zip(points[:-1], points[1:])]
    tp = [sum(links[0:i]) for i in range(len(links)+1)]
    xp, yp, zp, rp = points.T
    t = np.linspace(tp[0], tp[-1], num, endpoint=True)
    return np.array([np.interp(t, tp, xp),
                     np.interp(t, tp, yp),
                     np.interp(t, tp, zp),
                     np.interp(t, tp, rp)]).T



"""Implementation of CLI measure command."""

import os
import json
import math

import numpy as np

from treem import Morph, SWC

from treem.io import TreemEncoder
from treem.utils.geom import norm


def measure(args):
    """Computes morphometric features of the reconstruction."""
    # pylint: disable=invalid-name
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    # pylint: disable=cell-var-from-loop
    # pylint: disable=undefined-loop-variable
    types = args.type if args.type else SWC.TYPES
    ptmap = dict(zip(SWC.TYPES, ['soma', 'axon', 'dend', 'apic']))
    metric = dict()
    for reconstruction in args.file:
        morph = Morph(reconstruction)
        name = os.path.splitext(os.path.basename(reconstruction))[0]
        metric[name] = dict()

        for point_type in set(types).difference((SWC.SOMA,)):
            mdata = list()
            for sec in filter(lambda x: x[0].type() == point_type,
                              morph.root.sections()):
                head = sec[0]
                tail = sec[-1]
                seclen = morph.length(sec)
                chord = norm(tail.coord() - head.parent.coord())
                coords = morph.coords(sec)
                xmin, ymin, zmin = np.min(coords, axis=0)
                xmax, ymax, zmax = np.max(coords, axis=0)
                mdata.append(
                    [tail.degree(), head.order(), tail.breadth(),
                     seclen, chord / seclen,
                     morph.area(sec), morph.volume(sec),
                     morph.radii(sec).mean()*2,
                     xmin, xmax, ymin, ymax, zmin, zmax])
            if mdata:
                ndata = np.array(mdata)
                d = metric[name][ptmap[point_type]] = dict()
                d['degree'] = np.max(ndata[:, 0], axis=0).astype(int)
                d['order'] = np.max(ndata[:, 1], axis=0).astype(int)
                d['breadth'] = np.max(ndata[:, 2], axis=0).astype(int)
                d['nbranch'] = sum(1 for x in ndata[:, 0] if x > 1)
                d['nterm'] = sum(1 for x in ndata[:, 0] if x == 0)
                d['nstem'] = sum(1 for x in ndata[:, 1] if x == 1)
                d['length'] = np.sum(ndata[:, 3], axis=0)
                d['seclen'] = np.mean(ndata[:, 3], axis=0)
                d['contrac'] = np.mean(ndata[:, 4], axis=0)
                d['area'] = np.sum(ndata[:, 5], axis=0)
                d['volume'] = np.sum(ndata[:, 6], axis=0)
                d['diam'] = np.mean(ndata[:, 7], axis=0)
                d['xdim'] = (np.max(ndata[:, 9], axis=0)
                             - np.min(ndata[:, 8], axis=0))
                d['ydim'] = (np.max(ndata[:, 11], axis=0)
                             - np.min(ndata[:, 10], axis=0))
                d['zdim'] = (np.max(ndata[:, 13], axis=0)
                             - np.min(ndata[:, 12], axis=0))

        for point_type in set(types).intersection((SWC.SOMA,)):
            mdata = list()
            for sec in filter(lambda x: x[0].type() == point_type,
                              morph.root.sections()):
                if len(sec) > 1:
                    area = morph.area(sec)
                    volume = morph.volume(sec)
                else:
                    area = 4*math.pi*sec[0].radius()**2
                    volume = 4/3*math.pi*sec[0].radius()**3
                mdata.append([area, volume, morph.radii(sec).mean()*2])
            if mdata:
                ndata = np.array(mdata)
                d = metric[name][ptmap[point_type]] = dict()
                d['area'] = np.sum(ndata[:, 0], axis=0)
                d['volume'] = np.sum(ndata[:, 1], axis=0)
                d['diam'] = np.mean(ndata[:, 2], axis=0)
                d['xroot'], d['yroot'], d['zroot'] = morph.root.coord()

    if not args.out:
        for name in metric:
            print(name)
            for point_type in sorted(metric[name]):
                for feature in sorted(metric[name][point_type]):
                    print(f'{point_type} {feature:10s} '
                          f'{metric[name][point_type][feature]:>8g}')
            print()
    else:
        with open(args.out, 'w') as f:
            json.dump(metric, f, indent=4, sort_keys=True, cls=TreemEncoder)



"""Morphology reconstruction data structure."""

import math

from collections import deque

import numpy as np

from treem.tree import Tree
from treem.io import SWC, load_swc, save_swc
from treem.utils.geom import rotation_matrix, norm


class Node(Tree):
    """Morphology data storage."""

    def __init__(self, value=None):
        """Inits Node with value."""
        super(Node, self).__init__()
        self.v = value  # pylint: disable=invalid-name

    def __str__(self):
        """String representation of Node value."""
        return str(self.v)

    def walk(self, reverse=False):
        """Iterates through the tree nodes starting from the current node.

        Iteration is terminated when root is reached if reverse is
        True. If False, full tree is traversed downstream in pre-order.

        Args:
            reverse (bool): walk the tree in ascending order if True.

        Returns:
            sequence of tree nodes (generator object).
        """

        iterator = Tree.preorder if not reverse else Tree.ascendorder
        return iterator(self)

    def is_stem(self):
        """Returns True if node is a stem node."""
        return (not self.is_root() and self.parent.is_root()
                and self.type is not SWC.SOMA)

    def order(self):
        """Returns branch order (int). A primary neurite has order 1."""
        return (sum(1 for node in self.forks(iterator=Tree.ascendorder)) + 1
                if not self.is_root() else 0)

    def ident(self):
        """Returns node ID (int)."""
        return int(self.v[SWC.I])

    def parent_ident(self):
        """Returns node's parent ID (int)."""
        return int(self.v[SWC.P])

    def type(self):
        """Returns point type of the node (int)."""
        return int(self.v[SWC.T])

    def point(self):
        """Returns point data of the node (x,y,z,r) (NumPy ndarray[4])."""
        return self.v[SWC.XYZR]

    def coord(self):
        """Returns coordinates of the node (x,y,z) (NumPy ndarray[3])."""
        return self.v[SWC.XYZ]

    def radius(self):
        """Returns radius of the node (float)."""
        return self.v[SWC.R]

    def diam(self):
        """Returns diameter of the node (float)."""
        return 2*self.radius()

    def length(self):
        """Returns segment length at the node (float)."""
        # pylint: disable=invalid-name
        a = self.coord()
        b = self.parent.coord() if not self.is_root() else a
        return norm(a - b)

    def area(self):
        """Returns segment area at the node (float)."""
        # pylint: disable=invalid-name
        h = self.length()
        a = self.radius()
        b = self.parent.radius() if not self.is_root() else a
        return math.pi * (a + b) * math.sqrt((a - b) * (a - b) + h * h)

    def volume(self):
        """Returns segment volume at the node (float)."""
        # pylint: disable=invalid-name
        h = self.length()
        a = self.radius()
        b = self.parent.radius() if not self.is_root() else a
        return math.pi / 3.0 * (a * a + a * b + b * b) * h

    def section(self, reverse=False):
        """
        Iterates through the nodes of the section.

        Iteration starts with the current node and procedes until the
        end of the section.

        Args:
            reverse (bool): ascending order if True (defaults to False).

        Yields:
            sequence of nodes (generator object).
        """

        iterator = Tree.preorder if not reverse else Tree.ascendorder
        for node in iterator(self):
            yield node
            term = node if not reverse else node.parent
            if term.is_fork() or term.is_leaf() or term.is_root():
                break

    def sections(self):
        """Iterates through the sections in descending order.

           Iterations traverse entire branch starting with the current node.

           Yields:
               sequence of sections (generator object).
        """
        queue = deque((self.section(),))
        while queue:
            sec = list(queue.pop())
            node = sec[-1]
            siblings = deque(child.section() for child in node.siblings)
            queue.extend(reversed(siblings))
            yield sec


class Morph():
    """Neuron morphology representation."""

    def __init__(self, source=None, data=None):
        """Initializes Morph from source file or data.

        Args:
            source (str): source file (swc).
            data (NumPy ndarray): morphology data (N, 7).
        """
        self.data = None
        self.root = None
        self.nodes = list()
        if source:
            self.load(source)
        elif data is not None:
            data[0][SWC.P] = -1
            idmap = dict(enumerate([int(x[0]) for x in data], 1))
            idmap = {v: k for k, v in idmap.items()}
            idmap[-1] = -1
            for rec in data:
                rec[SWC.I] = idmap[int(rec[SWC.I])]
                rec[SWC.P] = idmap[int(rec[SWC.P])]
            self.load(data=data)

    def load(self, source=None, data=None):
        """Fill-in Morph from source file or data.

        Args:
            source (str): source file (swc).
            data (NumPy ndarray): morphology data (N, 7).
        """
        self.data = load_swc(source) if source else data
        for row in self.data:
            self.nodes.append(Node(row))
        self.root = self.nodes[0]
        for node in self.nodes[1:]:
            parent = node.parent_ident() - 1
            child = node.ident() - 1
            self.nodes[parent].add(self.nodes[child])

    def save(self, target):
        """Writes morphology to file (str)."""
        save_swc(target, self.data)

    def node(self, ident):
        """Returns node by it's ID."""
        return [node for node in self.root.walk() if node.ident() == ident][0]

    def stems(self):
        """Iterates through stem nodes.

        Returns:
            sequence of stem nodes (generator object).
        """
        return filter(lambda x: x.type() != SWC.SOMA, self.root.siblings)

    def coords(self, sec):
        """Returns reference to section coordinates."""
        first = sec[0].ident() - 1
        last = sec[-1].ident()
        block = slice(first, last)
        return self.data[block, SWC.XYZ]

    def radii(self, sec):
        """Returns reference to section radii."""
        first = sec[0].ident() - 1
        last = sec[-1].ident()
        block = slice(first, last)
        return self.data[block, SWC.RADII]

    def points(self, sec):
        """Returns reference to section data."""
        first = sec[0].ident() - 1
        last = sec[-1].ident()
        block = slice(first, last)
        return self.data[block, SWC.XYZR]

    def length(self, sec):  # pylint: disable=no-self-use
        """Returns section length (float)."""
        return sum(node.length() for node in sec)

    def area(self, sec):  # pylint: disable=no-self-use
        """Returns section area (float)."""
        return sum(node.area() for node in sec)

    def volume(self, sec):  # pylint: disable=no-self-use
        """Returns section volume (float)."""
        return sum(node.volume() for node in sec)

    def move(self, shift, node):  # pylint: disable=no-self-use
        """Shifts node coordinates by 3D vector (float[3])."""
        node.v[SWC.XYZ] += shift

    def translate(self, shift, node=None):
        """Shifts coordinates of the branch at the given node.

        Branch is traversed downstream from the given node.

        Args:
            shift (float[3]): translation vector.
            node (treem.Node): starting node (defaults to root).
        """
        node = node if node else self.root
        for sec in node.sections():
            points = self.coords(sec)
            points += shift

    def rotate(self, axis, angle, node=None):
        """Rotates branch at the node.

        Branch is traversed downstream from the given node.

        Args:
            axis (float[3]): rotation axis.
            angle (float): rotation angle in degrees.
            node (treem.Node): starting node (defaults to root).
        """
        node = node if node else self.root
        head = node.coord().copy()
        for sec in node.sections():
            points = self.coords(sec)
            first = sec[0].ident() - 1
            last = sec[-1].ident()
            block = slice(first, last)
            self.data[block, SWC.XYZ] = np.dot(rotation_matrix(axis, angle),
                                               points.T).T
        shift = head - node.coord()
        self.translate(shift, node)

    def copy(self, node=None):
        """Copies branch at the node (defaults to root)."""
        node = node if node else self.root
        data = np.array([x.v for x in node.walk()])
        return Morph(data=data)

    # Programming notes: 
    # 1) __renumber() changes internal container data;
    # 2) delete(), insert(), prune() and graft() desynchronize
    #    the data and the linked list;
    # 3) constructor Morph(data=new_data) updates the linked list.

    def __renumber(self):
        """Renumbers morphology nodes in tree traversal order."""
        data = np.array([x.v for x in self.root.walk()])
        idmap = dict(enumerate([int(x[0]) for x in data], 1))
        idmap = {v: k for k, v in idmap.items()}
        idmap[-1] = -1
        for rec in data:
            rec[SWC.I], rec[SWC.P] = idmap[rec[SWC.I]], idmap[rec[SWC.P]]
        self.data = data

    def delete(self, node):
        """Delete node."""
        siblings = node.parent.siblings
        index = siblings.index(node)
        siblings.pop(index)
        for child in node.siblings:
            node.parent.add(child)
            child.v[SWC.P] = node.parent.ident()
        self.__renumber()

    def insert(self, new_node, node):
        """Inserts new node before the given node."""
        siblings = node.parent.siblings
        index = siblings.index(node)
        siblings.pop(index)
        maxid = np.max(self.data[:, slice(SWC.I, SWC.I+1)]).astype(int)
        new_node.v[SWC.I] = maxid + 1
        new_node.v[SWC.P] = node.parent.ident()
        node.v[SWC.P] = new_node.ident()
        node.parent.add(new_node)
        new_node.add(node)
        self.__renumber()

    def prune(self, node):
        """Prunes branch at the given node."""
        siblings = node.parent.siblings
        index = siblings.index(node)
        siblings.pop(index)
        self.__renumber()

    def graft(self, tree, node=None):
        """Grafts tree at the given node (defaults to root)."""
        node = node if node else self.root
        maxid = np.max(self.data[:, slice(SWC.I, SWC.I+1)]).astype(int)
        tree.data[:, slice(SWC.I, SWC.P+1, SWC.P)] += maxid
        tree.data[0][SWC.P] = node.ident()
        self.data = np.append(self.data, tree.data, axis=0)
        node.add(tree.root)
        self.__renumber()
